/*
 * backbone-view-collection
 * https://github.com/PhilNehaev/backbone-view-collection
 *
 * Copyright (c) 2013 Philipp Nehaev
 * Licensed under the MIT license.
 */

/*global define*/

/** @module backbone-view-collection */

define([
    "backbone",
    "underscore"
], function(Backbone, _) {
    "use strict";

    return Backbone.View.extend(/** @lends ViewCollection */{
        /**
         * @example
         * var itemViews = new ViewCollection();
         * itemViews
         *  .setChunkSize(20)
         *  .add(views)
         *  .render();
         * itemViews
         *  .nextChunk()
         *  .render();
         *
         * @constructs
         * @param [options]
         * @param {Backbone.View[]|Backbone.View} [views] Call {@link #add} with this argument
         */
        initialize: function(options, views) {

            /**
             * List of view
             * @type {Backbone.View[]}
             * @private
             */
            this._views = [];

            /**
             * Length current chunks
             * @default 0
             * @type {Number}
             * @private
             */
            this._length = 0;

            /**
             * @type {DocumentFragment}
             * @private
             */
            this._fragment = document.createDocumentFragment();

            /**
             * For change use setCurrentChunk or nextChunk method
             * @default 1
             * @type {Number}
             * @private
             */
            this._currentChunk = 1;

            /**
             * For change use setChunkSize method
             * @default 999
             * @type {Number}
             * @private
             */
            this._chunkSize = 999;

            this._updateLimit();

            if (views) {

                this.add(views);
            }
        },

        /**
         * Name of a method for adding result
         * @type {String}
         */
        resultMethod: "html",

        /**
         * Add view
         * @param {Backbone.View[]|Backbone.View} views
         * @return {ViewCollection}
         */
        add: function(views) {

            if (_.isArray(views)) {

                this._views = this._views.concat(
                    this._preparingViews(views)
                );

            } else {

                this._views.push(
                    this._preparingView(views)
                );
            }

            return this;
        },

        /**
         * Destroy resources
         * @return {ViewCollection}
         */
        clear: function(options) {

            options = options || {};

            _.invoke(this._views, "remove");

            this._views = [];
            this._length = 0;

            if (!options.saveCurrentChunk) {

                this.setCurrentChunk(1);
            }

            this.$el.empty();

            return this;
        },

        /**
         * Update limit variable. Call after set currentChunk or chunkSize
         * @private
         * @return {ViewCollection}
         */
        _updateLimit: function() {

            /**
             * @type {Number}
             * @private
             */
            this._limit = this._currentChunk * this._chunkSize;

            return this;
        },

        /**
         * Set current chunk. Auto-call {@link #_updateLimit}
         * @param {Number} value
         * @return {ViewCollection}
         */
        setCurrentChunk: function(value) {

            this._currentChunk = value;
            this._updateLimit();

            return this;
        },

        /**
         * Next chunk. Auto-call {@link #_updateLimit}
         * @return {ViewCollection}
         */
        nextChunk: function() {

            this._currentChunk++;
            this._updateLimit();

            return this;
        },

        /**
         * Set chunk size. Auto-call {@link #_updateLimit}
         * @param {Number} value
         * @return {ViewCollection}
         */
        setChunkSize: function(value) {

            this._chunkSize = value;
            this._updateLimit();

            return this;
        },

        /**
         * @return {Number}
         */
        getChunkSize: function() {

            return this._chunkSize;
        },

        /**
         * @return {Number}
         */
        getLimit: function() {

            return this._limit;
        },

        /**
         * Return with refresh fragment
         * @return {ViewCollection}
         */
        render: function() {

            var i, view, currentViews = this._views;

            for (
                i = 0, view = this._views[i];
                view && i < this._limit;
                i++, view = this._views[i]
            ) {

                if (!view.isRendered) {

                    view.render();
                    view.isRendered = true;
                }
            }

            if (this._currentChunk && this._chunkSize) {

                currentViews = this._views.slice(0, this._limit);
            }

            _.each(
                _.pluck(currentViews, "el"),
                this._fragment.appendChild,
                this._fragment
            );

            this.$el[this.resultMethod](this._fragment);

            return this;
        },

        /**
         * @private
         * @param {Backbone.View} view
         * @return {ViewCollection}
         */
        _preparingView: function(view) {

            this._length++;

            _.extend(view, {

                ordinal: this._length
            });

            view.on("all", this.trigger, this);

            return view;
        },

        /**
         * @private
         * @param {Backbone.View[]} views
         * @return {Backbone.View[]}
         */
        _preparingViews: function(views) {

            return _.map(views, this._preparingView, this);
        },

        /**
         * @param {Function} callback
         * @param {Object} context
         * @return {ViewCollection}
         */
        each: function(callback, context) {

            _.each(this._views, callback, context);

            return this;
        }
    });
});